/****************************************************************
 * Debug routines
 *
 * Copyright (c) 1998, 1999
 *               Thomas Linder, Thomas.Linder@gmx.net
 *
 * $Id: debug.c,v 1.1.1.1 1999/08/15 19:54:41 tl Exp $
 ****************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] =
    "$Id: debug.c,v 1.1.1.1 1999/08/15 19:54:41 tl Exp $";

#if DEBUG

#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>

static int debuglevel = 0;

#define RCFILE			".debugrc"
#define FILE_MAXLEN		512

static FILE * DBGOUT = NULL;

int init_debug (const int dbglvl, char * tmpl)
{
    char dbgfile [FILE_MAXLEN];
    time_t now;
    
    debuglevel = 0;
    DBGOUT = NULL;
    
    if ( dbglvl <= 0 )
	return 0;

    if (tmpl == NULL)
    {
	DBGOUT = stderr;
    }
    else
    {
#if HAS_NPRINTF
	if ( strstr (tmpl, "%d") != NULL )
	{
	    if ( snprintf (dbgfile, FILE_MAXLEN, tmpl,
			   getpid ()) < 0 )
		return 1;
	}
	else
	{
	    char * t;

	    t = tmpl + strlen(tmpl);

	    strncpy (dbgfile, tmpl, FILE_MAXLEN);
	    
	    if ( snprintf (dbgfile, FILE_MAXLEN - strlen(tmpl),
			   t, getpid ()) < 0 )
		return 1;
	}
#else
	    /* Let's just hope... */
	if ( sprintf (dbgfile, tmpl, getpid ()) < 0 )
	    return 1;
#endif
	
	if ( (DBGOUT = fopen (dbgfile, "a")) == NULL )
	{
	    debuglevel = 0;
	    return 2;
	}
    }

    debuglevel = dbglvl;

    time (&now);
    fprintf (DBGOUT, "-------- level: %d, %s", debuglevel, ctime (&now));
    
    return 0;
}

void init_debug_rcenv (const char * rcline, const char * envvar)
{
    int level = 0;
    int where = 0;
    char * val = NULL;
    char * tmpl = NULL;

    if (rcline != NULL)
    {
#define MAXLINESIZE 512
	char line[MAXLINESIZE];
	char rcfile[FILE_MAXLEN] = "";
	char * t;
	FILE * fp;
	
	t = rcfile;
	
	if ((val = getenv ("HOME")) != NULL)
	{
	    strcpy (t, val);
	    t = &(t[strlen(rcfile)]);
	    *t++ = '/';
	    *t = '\0';
	}
	strncpy (t, RCFILE, FILE_MAXLEN - strlen (rcfile));
	if ((fp = fopen (rcfile, "r")) != NULL)
	{
	    val = NULL;
	    while (fgets (line, MAXLINESIZE, fp));
	    {
		t = line;
		while (isspace (*t))
		    t++;
		if (*t != '#')
		{
		    if (strncmp (rcline, line, strlen (rcline)) == 0)
		    {
			t += strlen (rcline);
			
			while (isspace (*t))
			    t++;
			
			val = t;
		    }
		}
	    }
	    fclose (fp);
	}
    }

    if (val == NULL)
    {
	if (envvar != NULL)
	{
	    if ((val = getenv (envvar)) == NULL)
		return;
	}
	else
	    return;
    }

    if ((val[0] == 'f') && (val[1] == ':'))
    {
	where = 1;
	val += 2;
    }

    if (val[0] == '\0')
	return;
    
    level = strtol (val, &tmpl, 0);

    if (where == 0)
    {
	if ((isspace(tmpl[0])) || (tmpl[0] == '\0'))
	    tmpl = NULL;
	else
	    return;
    }
    
    fprintf (stderr, "level %d debugging initialized!\n", level);
    
    (void) init_debug (level, tmpl);
}


int debug (const int level,
	   const char *fmt,
	   ...)
{
    va_list ap;
    int ret;

    if ( (!(level & debuglevel)) || (DBGOUT == NULL) )
	return -1;

    va_start(ap, fmt);
    ret = vfprintf (DBGOUT, fmt, ap);
    va_end(ap);
    return (ret);
}

void end_debug ()
{
    if ( (DBGOUT != NULL) && (DBGOUT != stderr) )
    {
	fclose (DBGOUT);
	DBGOUT = NULL;
    }

    return;
}

#endif
