/****************************************************************
 * Paint Shop Pro File Format Library
 * File I/O interface
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: fileio.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999 Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: fileio.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#define DEBUGENV psplibdbg
#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>

#include <psperrno.h>
int PSPErrno;

size_t
PSPread_int8 (FILE * fp,
	      psp_int8 * ptr,
	      size_t count)
{
    size_t total = 0;
    size_t current;
#if DEBUG
    int i;
    psp_int8 * p = ptr;
    fpos_t pos = -1;

    fgetpos (fp, &pos);
    
    debug (8192, "..time to read %lu bytes starting at %ld\n",
	   (unsigned long) count, (long) pos);
#endif
    
    while (count > 0)
    {
	current = fread ((char *) ptr, sizeof (psp_int8), count, fp);
	if (current <= 0)
	    break;
	count -= current;
	ptr += current;
	total += current;
    }
    
#if DEBUG    
    debug (8192, "read %lu bytes total\n", (unsigned long) total);

    debug (16384, "   ");
    for (i = 0; i < total; i++) {
	debug (16384, " %02x", p[i]);
    }
    debug (16384, "\n");
    
#endif

    return total;
}

size_t
PSPread_int16 (FILE * fp,
	       psp_int16 * ptr,
	       size_t count)
{
    size_t total;

    total = PSPread_int8 (fp, (psp_int8 *) ptr, count * 2);

    while (count--)
    {
	*ptr = psp_itoh16 (*ptr);
	ptr++;
    }

    debug (8192, "read %lu int16s total\n", (unsigned long) total / 2);

    return (total / 2);
}

size_t
PSPread_int32 (FILE * fp,
	       psp_int32 * ptr,
	       size_t count)
{
    size_t total;

    total = PSPread_int8 (fp, (psp_int8 *) ptr, count * 4);

    while (count--)
    {
	*ptr = psp_itoh32 (*ptr);
	ptr++;
    }

    debug (8192, "read %lu int32s total\n", (unsigned long) total / 4);

    return (total / 4);
}

size_t
PSPread_time_t (FILE * fp,
		time_t * ptr,
		size_t count)
{
    size_t total;

    total = PSPread_int8 (fp, (psp_int8 *) ptr, sizeof (time_t) * count);

    while (count--)
    {
	*ptr = psp_itohtime_t (*ptr);
	ptr++;
    }

    debug (8192, "read %lu time_t total\n",
	   (unsigned long) total / sizeof (time_t));

    return (total / sizeof (time_t));
}

#if X86_DOUBLE_IS_OK
size_t
PSPread_double (FILE * fp,
		double * ptr,
		size_t count)
{
    size_t total = 0;
    psp_int64 v;

    while (count--)
    {
	total += PSPread_int8 (fp, (psp_int8 *) &v, sizeof (psp_int64));
	debug (8, "0x%016Lx\n", v);
	*ptr++ = psp_itohdouble (v);
    }

    debug (8192, "read %lu double total\n",
	   (unsigned long) total / sizeof (psp_int64));

    return (total / sizeof (psp_int64));
}
#else /* X86_DOUBLE_IS_OK */
size_t
PSPread_double (FILE * fp,
		double * ptr,
		size_t count)
{
    size_t total = 0;
    psp_int64 v;

    while (count--)
    {
	total += PSPread_int8 (fp, (psp_int8 *) &v, sizeof (psp_int64));
	debug (8, "0x%016Lx\n", v);
	*ptr++ = 72.0;
    }

    debug (8192, "read %lu double total\n",
	   (unsigned long) total / sizeof (psp_int64));

    return (total / sizeof (psp_int64));
}
#endif /* X86_DOUBLE_IS_OK */

size_t
PSPwrite_int8 (FILE * fp,
	       const psp_int8 * ptr,
	       size_t count)
{
    size_t total = 0;
    size_t current;

    debug (8192, "..time to write %lu bytes\n", (unsigned long) count);

    while (count > 0)
    {
	current = fwrite ((char *) ptr, sizeof (psp_int8), count, fp);
	if (current <= 0)
	    break;
	count -= current;
	ptr += current;
	total += current;
    }
	
    debug (8192, "wrote %lu bytes total\n", (unsigned long) total);

    return (total);
}

size_t
PSPwrite_int16 (FILE * fp,
		const psp_int16 * ptr,
		size_t count)
{
    psp_int16 tmp;
    size_t total = 0;
    int i;

    for (i = 0; i < count; i++)
    {
	tmp = psp_htoi16 (ptr[i]);
	total += PSPwrite_int8 (fp, (psp_int8 *) &tmp, sizeof (psp_int16));
    }

    return (total / 2);
}

size_t
PSPwrite_int32 (FILE * fp,
		psp_int32 * ptr,
		size_t count)
{
    psp_int32 tmp;
    size_t total = 0;
    int i;

    for (i = 0; i < count; i++)
    {
	tmp = psp_htoi32 (ptr[i]);
	total += PSPwrite_int8 (fp, (psp_int8 *) &tmp, sizeof (psp_int32));
    }

    return (total / 4);
}
