/****************************************************************
 * Paint Shop Pro File Format Library
 * Hardware specific
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: hwspecific.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: hwspecific.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#include <debug.h>

#include <psperrno.h>
int PSPErrno;

/****************************************************************
 * BYTE ORDER INDEPENDENT read/write of basic datatypes
 ****************************************************************/

#if defined( HAVE_ENDIAN_H )
#include <endian.h>
#define ENDIAN			__BYTE_ORDER
#else /* HAS_ENDIAN_H */
#define __LITTLE_ENDIAN		1234
#define __BIG_ENDIAN		4321
#define __PDP_ENDIAN		3412
#endif /* HAS_ENDIAN_H */

#define __LITTLE_ENDIAN_DOUBLE	12345678
#define __BIG_ENDIAN_DOUBLE	87654321

#if ! defined( ENDIAN )
#error *** No information about byteorder
#endif /* ! defined( ENDIAN ) */

#if ! ENDIAN == __LITTLE_ENDIAN && ! ENDIAN == __LITTLE_ENDIAN
#error *** Target systems byte order not supported
#endif

/****************************************************************
 * Some basic checks
 */

#if SIZEOF_TIME_T != 4
#warning Oops! It really happened: sizeof(time_t) != 4
#endif

/* The read/write routines were inspired by the gimp xcf routines
 * but correspond more to the fread/fwrite data types
 */

#if ENDIAN == __LITTLE_ENDIAN
/* Little endian style like x86 doesn't need translation */
/* Convert from Intel to host byteorder */
inline psp_int16
psp_itoh16(psp_int16 intel16)
{
    return (intel16);
}
inline psp_int16
psp_htoi16(psp_int16 host16)
{
    return (host16);
}

inline psp_int32
psp_itoh32(psp_int32 intel32)
{
    return (intel32);
}
inline psp_int32
psp_htoi32(psp_int32 host32)
{
    return (host32);
}

/* This poses little problem at the moment, but what if the time_t type
 * is not a long integer value?
 */
inline psp_int32
psp_htoitime_t(time_t itime)
{
    return (itime);
}
inline time_t
psp_itohtime_t(psp_int32 itime)
{
    return (itime);
}

#endif /* ENDIAN == __LITTLE_ENDIAN */

#if ENDIAN == __BIG_ENDIAN
/* Big endian style is the way to do it ;) (like on 680xx - this is also
 * network byte order)
 */
inline psp_int16
psp_itoh16 (psp_int16 intel16)
{
    return (((intel16 & 0x00ff) << 8) |
	    ((intel16 & 0xff00) >> 8));
}
inline psp_int16
psp_htoi16 (psp_int16 host16)
{
    return (((host16 & 0x00ff) << 8) |
	    ((host16 & 0xff00) >> 8));
}

inline psp_int32
psp_itoh32 (psp_int32 intel32)
{
    return (((intel32 & 0x000000ff) << 24) |
	    ((intel32 & 0x0000ff00) << 8) |
	    ((intel32 & 0x00ff0000) >> 8) |
	    ((intel32 & 0xff000000) >> 24));
}
inline psp_int32
psp_htoi32 (psp_int32 host32)
{
    return (((host32 & 0x000000ff) << 24) |
	    ((host32 & 0x0000ff00) << 8) |
	    ((host32 & 0x00ff0000) >> 8) |
	    ((host32 & 0xff000000) >> 24));
}

/* This here is a problem again.  What if someday time_t should become
 * something other than 32 bits wide?  The PSP file format defines it as
 * a 4 byte value so just treat it like this?!
 * NOBODY will be happy with this...  You know the Year 2038 bug!
 */
inline psp_int32
psp_htoitime_t (time_t itime)
{
    return (((itime & 0x000000ff) << 24) |
	    ((itime & 0x0000ff00) << 8) |
	    ((itime & 0x00ff0000) >> 8) |
	    ((itime & 0xff000000) >> 24));
}
inline time_t
psp_itohtime_t (psp_int32 itime)
{
    return (((itime & 0x000000ff) << 24) |
	    ((itime & 0x0000ff00) << 8) |
	    ((itime & 0x00ff0000) >> 8) |
	    ((itime & 0xff000000) >> 24));
}

#endif /* ENDIAN == __BIG_ENDIAN */

#if DOUBLES_ARE_OK
/* FIXME! Floats are not integers :( */
/* Problem is, this is even more host dependant than integer byteorders
 * are, as not everyone might use the IEEE floating point format for
 * these double precision bytes.
 * If new byte-orders are supported, tests have to be added to configure.in
 */
#if DOUBLE_ENDIAN == __LITTLE_ENDIAN_DOUBLE
inline double
psp_itohdouble (psp_int64 dbl)
{
    double val;
    unsigned char * t;
    unsigned char * s;

    t = (unsigned char *) &val;
    s = (unsigned char *) &dbl;

    debug (8, "itohdouble: %02x %02x %02x %02x  %02x %02x %02x %02x -> ",
	   s[0], s[1], s[2], s[3], s[4], s[5], s[6], s[7]);

    *t++ = *s++, *t++ = *s++, *t++ = *s++, *t++ = *s++,
    *t++ = *s++, *t++ = *s++, *t++ = *s++, *t++ = *s++;

    debug (8, "%02x %02x %02x %02x  %02x %02x %02x %02x -> %f\n",
	   t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], val);
    return val;
}
inline psp_int64
psp_htoidouble (double dbl)
{
    return dbl;
}

#else /* DOUBLE_ENDIAN == __LITTLE_ENDIAN_DOUBLE */
#if DOUBLE_ENDIAN == __BIG_ENDIAN_DOUBLE
inline double
psp_itohdouble (psp_int64 dbl)
{
    double val;
    unsigned char * t;
    unsigned char * s;

    t = (unsigned char *) &val + 7;
    s = (unsigned char *) &dbl;

    debug (8, "itohdouble: %02x %02x %02x %02x  %02x %02x %02x %02x -> ",
	   s[0], s[1], s[2], s[3], s[4], s[5], s[6], s[7]);

    *t-- = *s++, *t-- = *s++, *t-- = *s++, *t-- = *s++,
    *t-- = *s++, *t-- = *s++, *t-- = *s++, *t-- = *s++;

    debug (8, "%02x %02x %02x %02x  %02x %02x %02x %02x -> %f\n",
	   t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], val);

    return val;
}

inline psp_int64
psp_htoidouble (double dbl)
{
    psp_int64 val;
    unsigned char * t;
    unsigned char * s;

    t = (unsigned char *) &val;
    s = (unsigned char *) &dbl + 7;

    *t++ = *s--, *t++ = *s--, *t++ = *s--, *t++ = *s--;
    *t++ = *s--, *t++ = *s--, *t++ = *s--, *t++ = *s--;

    return val;
}
#else /* DOUBLE_ENDIAN == __BIG_ENDIAN_DOUBLE */
#error *** Unsupported double byteorder
#endif /* DOUBLE_ENDIAN == __BIG_ENDIAN_DOUBLE */

#endif /* DOUBLE_ENDIAN == __LITTLE_ENDIAN_DOUBLE */
#else
#warning *** Double value import/export not supported for this system
#endif /* DOUBLES_ARE_OK */
