/****************************************************************
 * Paint Shop Pro File Format Library
 * Main interface
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: psp.c,v 1.1.1.1 1999/08/15 19:54:41 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid[] = "$Id: psp.c,v 1.1.1.1 1999/08/15 19:54:41 tl Exp $";
static char const copyright[] =
    "Copyright (c) 1999  Thomas Linder, e8826423@student.tuwien.ac.at";    

#include "psp.h"
#include <pspproto.h>

#include <debug.h>

#include <psperrno.h>
int PSPErrno;

#if 0
const char PSP_Signature [32] = "Paint Shop Pro Image File\n\x1a\0\0\0\0\0";
const char PSP_BlockSig [4] = "~BK\0";
const char PSP_SubBlockSig [4] = "~FL\0";
#endif

const WORD PSP_Major = 3;
const WORD PSP_Minor = 0;


/****************************************************************
 * LIBRARY ROUTINES
 ****************************************************************/

/****************************************************************
 * PSPOpen
 *
 * Open a PSP file, like the stdio fopen but returns a pointer to
 * the PSP struct instead.
 */

PSP *
PSPOpen (const char * path,
	 const char * mode)
{
    PSP * psp;
    char fmode[3] = "";

    init_debug_rcenv ("gimp-psp-debug", "GIMP_PSP_DEBUG");

    debug (256, "=> PSPOpen (const char * \"%s\", const char * \"%s\");\n",
	   path, mode);

    PSPErrno = PSPENOERR;

    if ((psp = (PSP *) PSPLIBMalloc (sizeof (PSP))) == NULL)
	return NULL;

    psp->fp = NULL;
    psp->extended = NULL;
/* CHANGED    psp->creator = NULL; */
    psp->thumbnail = NULL;
    psp->palette = NULL;
    psp->layer = NULL;
    psp->selection = NULL;
    psp->alpha = NULL;
    psp->layernum = -1;
    psp->channelnum = -1;
    psp->inselection = FALSE;
    psp->inalpha = FALSE;
    psp->rle.runtype = 0;
    psp->rle.runcount = 0;
    psp->rle.lastread = 0;

    if (mode[0] == 'r')
    {
	strcpy (fmode, "rb");
    }
    else if (mode[0] == 'w')
    {
	strcpy (fmode, "wb");
    }
    else
    {
	PSPErrno = PSPEINVAL;
	return (NULL);
    }

    if ( mode[1] != '\0' && (mode[1] != 'b' || mode[2] != '\0') )
    {
	PSPErrno = PSPEINVAL;
	return (NULL);
    }

    if ((psp->fp = fopen (path, fmode)) == NULL)
    {
	PSPClose (psp);
	return NULL;
    }

    debug (2, "Opened with file pointer is 0x%08lx\n", psp->fp);
    
	/* Now let's check if this is really a PSP, but only if it
	 * won't be created/overwritten
	 */
	/* Assume, fopen already did a check of the mode flags */
    if (mode[0] == 'w')
    {
	if (PSPwrite_int8 (psp->fp, PSP_Signature, sizeof PSP_Signature)
	    != sizeof PSP_Signature)
	{
	    debug (16, "Unable to write PSP File Signature!\n");
	    PSPClose (psp);
	    unlink (path);
	    return (NULL);
	}
    }
    else if (mode [0] == 'r')
    {
	if (PSPreadinfo (psp) == NULL)
	{
	    PSPClose (psp);
	    return (NULL);
	}
    }
    else
    {
	PSPClose (psp);
	PSPErrno = PSPEINVAL;
	return (NULL);
    }

    return psp;
}

/****************************************************************
 * PSPClose
 *
 * Like stdio's fclose but takes a PSP struct pointer as
 * argument
 */

int
PSPClose (PSP * psp)
{
    int ret = 0;
    int i, j;

    debug (256, "=> PSPClose (PSP * 0x%08lx);\n", psp);

    if (psp->extended != NULL)
    {
	for (i = 0; psp->extended[i] != NULL; i++)
	    PSPLIBFree (psp->extended[i]);
	PSPLIBFree (psp->extended);
    }

/* CHANGED
    if (psp->creator != NULL)
    {
	for (i = 0; psp->creator[i] != NULL; i++)
	    PSPLIBFree (psp->creator[i]);
	PSPLIBFree (psp->creator);
    }
*/
    
    if (psp->thumbnail != NULL)
	PSPLIBFree (psp->thumbnail);

    if (psp->palette != NULL)
	PSPLIBFree (psp->palette);

    if (psp->layer != NULL)
    {
	for (i = 0; (psp->layer[i] != NULL) && (i < psp->general.layers); i++)
	{
	    if (psp->layer[i]->channel != NULL)
	    {
		for (j = 0; (psp->layer[i]->channel[j] != NULL)
			 && (j < psp->layer[i]->info.channels); j++)
		{
		    PSPLIBFree (psp->layer[i]->channel[j]);
		}
		PSPLIBFree (psp->layer[i]->channel);
	    }
		
	    PSPLIBFree (psp->layer[i]);
	}
	
	PSPLIBFree (psp->layer);
    }

    if (psp->selection != NULL)
    {
	if (psp->selection->channel != NULL)
	{
	    for (j = 0; (psp->selection->channel[j] != NULL)
		     && (j < psp->selection->channels); j++)
	    {
		PSPLIBFree (psp->selection->channel[j]);
	    }
	    PSPLIBFree (psp->selection->channel);
	}
	PSPLIBFree (psp->selection);
    }
    
    if (psp->alpha != NULL)
    {
	for (i = 0; (psp->alpha[i] != NULL) && (i < psp->alphacount); i++)
	{
	    if (psp->alpha[i]->channel != NULL)
	    {
		for (j = 0; (psp->alpha[i]->channel[j] != NULL)
			 && (j < psp->alpha[i]->info.channels); j++)
		{
		    PSPLIBFree (psp->alpha[i]->channel[j]);
		}
		PSPLIBFree (psp->alpha[i]->channel);
	    }
		
	    PSPLIBFree (psp->alpha[i]);
	}
	
	PSPLIBFree (psp->alpha);
    }

    if (psp->fp != NULL)
	ret = fclose (psp->fp);

	/* note that free()/PSPFree() return void */
    PSPLIBFree (psp);

    debug (2, "Closed with returncode %d\n", ret);

    return (ret);
}

/****************************************************************
 * PSPGetField
 *
 * Read information from the opened PSP file
 */
int
PSPGetField (PSP * psp, pspf_t field, void * value)
{
    long block;
    int ret = 0;

    debug (256, "=> PSPGetField (PSP * 0x%08lx, pspf_t 0x%08lx, "
	   "void * 0x%08lx);\n", psp, field, value);

    block = field & 0xffff0000;
    
    PSPErrno = PSPENOERR;
    switch (block)
    {
      case PSPB_GENERAL:
	  ret = PSPGetGeneralField (psp, field, value);
	  break;
      case PSPB_CREATOR:
	  ret = PSPGetCreatorField (psp, field, value);
	  break;
      case PSPB_LAYER:
	  ret = PSPGetLayerField (psp, field, value);
	  break;
      case PSPB_CHANNEL:
	  ret = PSPGetChannelField (psp, field, value);
	  break;
      case PSPB_CMAP:
	  ret = PSPGetCMapField (psp, field, value);
	  break;
      case PSPB_SELECTION:
	  ret = PSPGetSelectionField (psp, field, value);
	  break;
      default:
	  PSPErrno = PSPEINVAL;
    }
    
    return ret;
}

/****************************************************************
 * PSPMalloc
 *
 * PSP wrapper to the standard C library function malloc ()
 */
void *
PSPMalloc (size_t size)
{
    void * ptr;

    ptr = malloc (size);
    
    debug (4096, ".. allocated %ld bytes at %08lx\n",
	   (long) size, (long) ptr);

    return (ptr);
}

/****************************************************************
 * PSPCalloc
 *
 * PSP wrapper to the standard C library function calloc ()
 */
void *
PSPCalloc (size_t nmemb, size_t size)
{
    void * ptr;

    ptr = calloc (nmemb, size);

    debug (4096, ".. allocated %ld x %ld bytes at %08lx\n",
	   (long) nmemb, (long) size, (long) ptr);

    return (ptr);
}

/****************************************************************
 * PSPFree
 *
 * PSP wrapper to the standard C library function free ()
 */
void
PSPFree (void * ptr)
{
    debug (4096, ".. deallocate memory at %08lx\n", (long) ptr);

    free (ptr);
}

/****************************************************************
 * PSPStrError
 *
 *
 */
char *
PSPStrError (int PSPErrnum)
{
    static char err [] = "this function is not yet implemented";
    
    return (err);
}

/****************************************************************
 * PSPDebug
 *
 * initialize debugging
 */

int
PSPDebug (const int debuglevel, const char * tmpl)
{
    return init_debug (debuglevel, tmpl);
};
