/****************************************************************
 * Paint Shop Pro File Format Library
 * Alpha channel handling
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: pspalpha.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: pspalpha.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include <psperrno.h>
int PSPErrno;


struct PSPAlphaInfo **
PSPReadAlphaInfo (PSP * psp,
		  DWORD firstsize,
		  DWORD blocksize)
{
    struct PSPAlphaInfo * this;
    fpos_t endpos;
    fpos_t nextpos;
    fpos_t pos;
    B_ARRAY signature[4];
    WORD id;
    DWORD size;
    DWORD subblocksize;
    int i;

    if (fgetpos (psp->fp, &endpos) != 0)
	return NULL;
    endpos += blocksize;

    if (PSPread_int16 (psp->fp, &psp->alphacount, 1) != 1)
    {
	debug (16, "Unable to read alpha channel count (EOF?)\n");
	return NULL;
    }
		       
    if ((psp->alpha = PSPLIBCalloc (psp->alphacount,
				    sizeof (struct PSPAlphaInfo *)))
	== NULL)
    {
	debug (8, "Unable to allocate memory for Alpha Channel Info Array\n");
	return NULL;
    }

    for (i = 0; i < psp->alphacount ; i++)
    {
	if (fgetpos (psp->fp, &pos) != 0)
	{
	    debug (16, "Either some error occured in fgetpos or the "
		   "end-of-file has been reached before all data has "
		   "been read");
	    return NULL;
	}
	
	if (pos >= endpos)
	{
	    debug (16, "Something wrong here, not all alpha channels read but"
		   "end of block already reached!\n");
	    return NULL;
	}
	
	if (PSPread_int8 (psp->fp, signature, sizeof (signature))
	    != sizeof (signature)
	    || PSPread_int16 (psp->fp, &id, 1) != 1
	    || PSPread_int32 (psp->fp, &size, 1) != 1
	    || PSPread_int32 (psp->fp, &subblocksize, 1) != 1
	    )
	{
	    if (! feof (psp->fp))
		return NULL;
	    debug (16, "Error reading next alpha channel sub-block (EOF?)\n");
	    return NULL;
	}
	
	if (memcmp (PSP_BlockSig, signature, sizeof (PSP_BlockSig)) != 0)
	{
	    debug (16, "No alpha channel sub-block signature found where "
		   "expected! (%c%c%c%c)\n", signature[0],
		   signature[1], signature[2], signature[3]);
	    PSPErrno = PSPENOPSP;
	    return NULL;
	}

	if (id != PSP_ALPHA_CHANNEL_BLOCK)
	{
	    debug (16, "Wrong block identifier %hu where PSP_LAYER_BLOCK "
		   "(%hu) expected!\n", id, PSP_ALPHA_CHANNEL_BLOCK);
	    PSPErrno = PSPENOPSP;
	    return NULL;
	}

	if (fgetpos (psp->fp, &nextpos) != 0)
	    return NULL;

	nextpos += subblocksize;

	if ((psp->alpha[i] = PSPLIBMalloc (sizeof (struct PSPAlphaInfo)))
	    == NULL)
	{
	    debug (8, "Unable to allocate memory for Alpha Channel "
		   "%d Info \n", i);
	    return NULL;
	}
	
	this = psp->alpha[i];

	if (PSPread_int8 (psp->fp, this->info.name, PSP_ALPHAINFONAMELEN) !=
	    PSP_ALPHAINFONAMELEN
	    || PSPread_int32 (psp->fp, &this->info.rect.xmin, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.rect.ymin, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.rect.xmax, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.rect.ymax, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.saved_rect.xmin, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.saved_rect.ymin, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.saved_rect.xmax, 1) != 1
	    || PSPread_int32 (psp->fp, &this->info.saved_rect.ymax, 1) != 1
	    || PSPread_int16 (psp->fp, &this->info.bitmaps, 1) != 1
	    || PSPread_int16 (psp->fp, &this->info.channels, 1) != 1)
	{
	    debug (16, "Error reading alpha channel information\n");
	    return NULL;
	}

	debug (32, "Alpha channel %d -> name: %.256s\n", i, this->info.name);
	debug (32, "Alpha channel %d -> rect: (%ld, %ld), (%ld, %ld)\n",
	       i, this->info.rect.xmin, this->info.rect.ymin,
	       this->info.rect.xmax, this->info.rect.ymax);
	debug (32, "Alpha channel %d -> saved rect: (%ld, %ld), (%ld, %ld)\n",
	       i, this->info.saved_rect.xmin, this->info.saved_rect.ymin,
	       this->info.saved_rect.xmax, this->info.saved_rect.ymax);
	
	debug (32, "Alpha channel %d -> bitmap count: %hu\n",
	       i, this->info.bitmaps);
	debug (32, "Alpha channel %d -> channel count: %hu\n",
	       i, this->info.channels);
	
	    /* Alpha channel is always 8 bit "greyscale" */
	this->scanlinebytes = this->info.saved_rect.xmax
	    - this->info.saved_rect.xmin;
	this->scanlinealign = this->scanlinebytes;
	this->scanlinealign += 3;
	this->scanlinealign &= 0xfffffffc;
	
	debug (32, "Alpha channel %d -> scanline bytecount: %lu\n",
	       i, this->scanlinebytes);
	debug (32, "Alpha channel %d -> scanline aligned bytecount: %lu\n",
	       i, this->scanlinealign);
	
#if 0
	if ( (bmcount != 1) || (chcount != 1) )
	{
	    debug (16, "Unsupported selection\n");
	    return NULL;
	}
#endif

	if (PSPReadChannelInfo (psp->fp, size, subblocksize,
				&this->channel, this->info.channels)
	    == NULL)
	{
	    debug (16, "Unable to read alpha channel content\n");
	    return NULL;
	}
    }
    
    return psp->alpha;
}

int
PSPGetAlphaField (PSP * psp,
		  pspf_t field,
		  void * value)
{
    struct PSPAlphaInfo * alpha;

    if (psp->alpha == NULL)
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    if ((psp->layernum < 0) || (psp->layernum >= psp->alphacount))
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    alpha = psp->alpha[psp->layernum];
	
    switch (field)
    {
      case PSPF_ALPHANAME:
	  if ((*((char **) value) =
	       PSPMalloc (strlen (alpha->info.name) + 1)) == NULL)
	      return 0;
	  strcpy (*((char **) value), alpha->info.name);
	  break;
      case PSPF_ALPHAEXTENT:
	  *((psprect_t *) value) = alpha->info.rect;
	  break;
      case PSPF_ALPHADATAEXTENT:
	  *((psprect_t *) value) = alpha->info.saved_rect;
	  break;
      case PSPF_ALPHABITMAPCOUNT:
	  *((pspword_t *) value) = alpha->info.bitmaps;
	  break;
      case PSPF_ALPHACHANNELCOUNT:
	  *((pspword_t *) value) = alpha->info.channels;
	  break;
      case PSPF_ALPHASCANLINEBYTES:
	  *((pspdword_t *) value) = alpha->scanlinebytes;
	  break;
      case PSPF_ALPHASCANLINEALIGN:
	  *((pspdword_t *) value) = alpha->scanlinealign;
	  break;
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }
    return 1;
}
