/****************************************************************
 * Paint Shop Pro File Format Library
 * Channel handling
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: pspchannel.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: pspchannel.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include <psperrno.h>
int PSPErrno;

#define RLESLOW 1

struct PSPChannelInfo ***
PSPReadChannelInfo (FILE * fp, DWORD firstsize, DWORD blocksize,
		    struct PSPChannelInfo *** channel, WORD numchannels)
{
    fpos_t pos = 0;
    fpos_t nextpos;
    struct PSPChannelInfo * this;
    B_ARRAY signature[4];
    WORD id;
    DWORD infosize;
    DWORD fullsize;
    WORD i;

	/* calloc clears memory, which we use to our advantage later */
    if ((*channel = PSPLIBCalloc (numchannels,
				  sizeof (struct PSPChannelInfo *))) == NULL)
    {
	debug (8, "Unable to allocate memory for Channel Info Array\n");
	return NULL;
    }

    for (i = 0; i < numchannels; i++)
    {
	    /* dummy
	this = (*channel)[i];
	    */
	
	if ((this = PSPLIBMalloc (sizeof (struct PSPChannelInfo))) == NULL)
	{
	    debug (8, "Unable to allocate memory for Channel Info\n");
	    return NULL;
	}
	(*channel)[i] = this;
	
	if (PSPread_int8 (fp, signature, sizeof (signature))
	    != sizeof (signature)
	    || PSPread_int16 (fp, &id, 1) != 1
	    || PSPread_int32 (fp, &infosize, 1) != 1
	    || PSPread_int32 (fp, &fullsize, 1) != 1
	    )
	{
	    if (! feof (fp))
		return NULL;
	    debug (16, "Error reading next layer sub-block (short file)!\n");
	    return NULL;
	}
	
	if (memcmp (PSP_BlockSig, signature, sizeof (PSP_BlockSig)) != 0)
	{
	    debug (16, "No channel block signature found where expected! "
		   "(%c%c%c%c)\n", signature[0],
		   signature[1], signature[2], signature[3]);
	    PSPErrno = PSPENOPSP;
	    return NULL;
	}
	
	if (id != PSP_CHANNEL_BLOCK)
	{
	    debug (16, "Wrong block identifier %hu where PSP_LAYER_BLOCK "
		   "(%hu) expected!\n", id, PSP_CHANNEL_BLOCK);
	    return NULL;
	}

	if (fgetpos (fp, &pos) != 0)
	{
	    debug (16, "Unable to get file position!\n");
	    return NULL;
	}
	nextpos = pos + fullsize;
	pos += infosize;
	this->startpos = pos;
	    
	debug (64, "  channel: %d, size: %lu (->%ld)\n", i,
	       fullsize, nextpos);

	if (PSPread_int32 (fp, &this->info.compress_size, 1) != 1
	    || PSPread_int32 (fp, &this->info.size, 1) != 1
	    || PSPread_int16 (fp, &this->info.bitmap_type, 1) != 1
	    || PSPread_int16 (fp, &this->info.channel_type, 1) != 1)
	{
	    debug (16, "Error reading channel information\n");
	    return NULL;
	}

#if DEBUG
	debug (32, "Channel %d -> compressed size: %lu\n", i, (long) this->info.compress_size);
	debug (32, "Channel %d -> size: %lu\n", i, (long) this->info.size);
	debug (32, "Channel %d -> bitmap type: ", i);
	
	switch (this->info.bitmap_type)
	{
	  case PSP_DIB_IMAGE:
	      debug (32, "layer color\n");
	      break;
	  case PSP_DIB_TRANS_MASK:
	      debug (32, "layer transparency mask\n");
	      break;
	  case PSP_DIB_USER_MASK:
	      debug (32, "layer user mask\n");
	      break;
	  case PSP_DIB_SELECTION:
	      debug (32, "selection mask\n");
	      break;
	  case PSP_DIB_ALPHA_MASK:
	      debug (32, "alpha channel mask\n");
	      break;
	  case PSP_DIB_THUMBNAIL:
	      debug (32, "thumbnail\n");
	      break;
	  default:
	      debug (32, "unknown. Need more information\n");
	}
	debug (32, "Channel %d -> channel type: ", i);
	switch (this->info.channel_type)
	{
	  case PSP_CHANNEL_COMPOSITE:
	      debug (32, "indexed or grayscale\n");
	      break;
	  case PSP_CHANNEL_RED:
	      debug (32, "red\n");
	      break;
	  case PSP_CHANNEL_GREEN:
	      debug (32, "green\n");
	      break;
	  case PSP_CHANNEL_BLUE:
	      debug (32, "blue\n");
	      break;
	  default:
	      debug (32, "unknown. Need more information\n");
	}
	debug (32, "Channel %d -> startpos: %lu\n", i, (long) this->startpos);
#endif

	if (fsetpos (fp, &nextpos) != 0)
	{
	    debug (16, "Unable to set file position!\n");
	    return NULL;
	}
    }
    
    return channel;
}

static int
PSPSearchChannel (struct PSPLayerInfo * layer,
		  int bitmap_type,
		  int channel_type)
{
    long chnum;
    struct PSP_ChannelInformationChunk chinfo;

    for (chnum = 0; chnum < layer->info.channels; chnum++)
    {
	chinfo = layer->channel[chnum]->info;
	if ((chinfo.bitmap_type == bitmap_type)
	    && (chinfo.channel_type == channel_type))
	    return chnum;
    }
	/* search failed */
    return -1;
}

static int
PSPSetSelectionChannel (PSP * psp,
			pspcht_t channeltype)
{
    if (psp->selection == NULL)
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }
    
    switch (channeltype)
    {
      case PSP_CHANNELSELECTIONFIRST:
	  psp->channelnum = 0;
	  break;
      case PSP_CHANNELSELECTIONLAST:
	  psp->channelnum = psp->selection->channels - 1;
	  break;
      case PSP_CHANNELSELECTIONNEXT:
	  if (psp->channelnum + 1 >= psp->selection->channels)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum++;
	  break;
      case PSP_CHANNELSELECTIONPREV:
	  if (psp->channelnum <= 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum--;
	  break;
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }

    if (fsetpos (psp->fp,
		 &(psp->selection->channel[psp->channelnum]->startpos))
	!= 0) 
    {
	debug (16, "Unable to set file position for channel operations!\n");
	return 0;
    }

    psp->inalpha = FALSE;
    psp->inselection = TRUE;
    
    return 1;
}

static int
PSPSetAlphaChannel (PSP * psp,
		    pspcht_t channeltype)
{
    struct PSPAlphaInfo * alpha;

    if ((psp->alpha == NULL) || (psp->inalpha == FALSE))
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    if ((psp->layernum < 0) || (psp->layernum >= psp->alphacount))
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    alpha = psp->alpha[psp->layernum];
    
    switch (channeltype)
    {
      case PSP_CHANNELALPHAFIRST:
	  psp->channelnum = 0;
	  break;
      case PSP_CHANNELALPHALAST:
	  psp->channelnum = alpha->info.channels - 1;
	  break;
      case PSP_CHANNELALPHANEXT:
	  if (psp->channelnum + 1 >= alpha->info.channels)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum++;
	  break;
      case PSP_CHANNELALPHAPREV:
	  if (psp->channelnum <= 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum--;
	  break;
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }

    if (fsetpos (psp->fp,
		 &(alpha->channel[psp->channelnum]->startpos))
	!= 0)
    {
	debug (16, "Unable to set file position for channel operations!\n");
	return 0;
    }

    psp->inselection = FALSE;
    
    return 1;
}

int
PSPSetChannel (PSP * psp,
	       pspcht_t channeltype)
{
    struct PSPLayerInfo * layer;
    long chnum;
    
    PSPErrno = PSPENOERR;

    if ((psp->inalpha) || (channeltype <= PSP_CHANNELALPHAOFFSET))
    {
	return PSPSetAlphaChannel (psp, channeltype);
	return 0;
    }
    if (channeltype <= PSP_CHANNELSELECTIONOFFSET)
    {
	return PSPSetSelectionChannel (psp, channeltype);
    }
    
    if ((psp->layernum < 0) || (psp->layernum >= psp->general.layers))
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    layer = psp->layer[psp->layernum];

    switch (channeltype)
    {
	    /* first the easy stuff */
      case PSP_CHANNELFIRST:
	  psp->channelnum = 0;
	  break;
      case PSP_CHANNELLAST:
	  psp->channelnum = layer->info.channels - 1;
	  break;
      case PSP_CHANNELNEXT:
	  if (psp->channelnum + 1 >= layer->info.channels)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum++;
	  break;
      case PSP_CHANNELPREV:
	  if (psp->channelnum <= 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum--;
	  break;
      case PSP_CHANNELCOMPOSITE:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_IMAGE,
					 PSP_CHANNEL_COMPOSITE)) < 0)
	  {
	      if ((chnum = PSPSearchChannel (layer, PSP_DIB_THUMBNAIL,
					     PSP_CHANNEL_COMPOSITE)) < 0)
	      {
		  PSPErrno = PSPEINVAL;
		  return 0;
	      }
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELRED:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_IMAGE,
					 PSP_CHANNEL_RED)) < 0)
	  {
	      if ((chnum = PSPSearchChannel (layer, PSP_DIB_THUMBNAIL,
					     PSP_CHANNEL_RED)) < 0)
	      {
		  PSPErrno = PSPEINVAL;
		  return 0;
	      }
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELGREEN:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_IMAGE,
					 PSP_CHANNEL_GREEN)) < 0)
	  {
	      if ((chnum = PSPSearchChannel (layer, PSP_DIB_THUMBNAIL,
					     PSP_CHANNEL_GREEN)) < 0)
	      {
		  PSPErrno = PSPEINVAL;
		  return 0;
	      }
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELBLUE:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_IMAGE,
					 PSP_CHANNEL_BLUE)) < 0)
	  {
	      if ((chnum = PSPSearchChannel (layer, PSP_DIB_THUMBNAIL,
					     PSP_CHANNEL_BLUE)) < 0)
	      {
		  PSPErrno = PSPEINVAL;
		  return 0;
	      }
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELTRANSMASK:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_TRANS_MASK,
					 PSP_CHANNEL_COMPOSITE)) < 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELUSERMASK:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_USER_MASK,
					 PSP_CHANNEL_COMPOSITE)) < 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELSELECTION:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_SELECTION,
					 PSP_CHANNEL_COMPOSITE)) < 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum = chnum;
	  break;
      case PSP_CHANNELALPHA:
	  if ((chnum = PSPSearchChannel (layer, PSP_DIB_ALPHA_MASK,
					 PSP_CHANNEL_COMPOSITE)) < 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum = chnum;
	  break;
      default:
	  if ((channeltype < 0) || (channeltype >= layer->info.channels))
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->channelnum = channeltype;
    }

    if (fsetpos (psp->fp,
		 &(layer->channel[psp->channelnum]->startpos))
	!= 0) 
    {
	debug (16, "Unable to set file position for channel operations!\n");
	return 0;
    }

    psp->inselection = FALSE;

    return 1;
}

int
PSPNextChannel (PSP * psp)
{
    if (psp->inalpha)
	return (PSPSetAlphaChannel (psp, PSP_CHANNELALPHANEXT));
    else if (psp->inselection)
	return (PSPSetSelectionChannel (psp, PSP_CHANNELSELECTIONNEXT));
    else
	return (PSPSetChannel (psp, PSP_CHANNELNEXT));
}

static pspcht_t
PSPGetChannelType (PSP * psp)
{
    struct PSP_ChannelInformationChunk chinfo;

    PSPErrno = PSPENOERR;

    if (psp->inalpha)
    {
	chinfo = psp->alpha[psp->layernum]->channel[psp->channelnum]->info;
    }
    else if (psp->inselection)
    {
	chinfo = psp->selection->channel[psp->channelnum]->info;
    }
    else
    {
	if ((psp->layernum < 0) ||
	    (psp->layernum >= psp->general.layers))
	{
	    PSPErrno = PSPEINVAL;
	    return PSP_CHANNELNONE;
	}    

	chinfo = psp->layer[psp->layernum]->channel[psp->channelnum]->info;
    }

    if (chinfo.bitmap_type == PSP_DIB_IMAGE)
    {
	switch (chinfo.channel_type)
	{
	  case PSP_CHANNEL_COMPOSITE:
	      return PSP_CHANNELCOMPOSITE;
	  case PSP_CHANNEL_RED:
	      return PSP_CHANNELRED;
	  case PSP_CHANNEL_GREEN:
	      return PSP_CHANNELGREEN;
	  case PSP_CHANNEL_BLUE:
	      return PSP_CHANNELBLUE;
	  default:
	      PSPErrno = PSPENOPSP;
	      return PSP_CHANNELNONE;
	}
    }
    else if (chinfo.bitmap_type == PSP_DIB_THUMBNAIL)
    {
	PSPErrno = PSPENIMPL;
	return PSP_CHANNELNONE;
	    /* The following is not reached */
	switch (chinfo.channel_type)
	{
	  case PSP_CHANNEL_COMPOSITE:
	      return PSP_CHANNELCOMPOSITE;
	  case PSP_CHANNEL_RED:
	      return PSP_CHANNELRED;
	  case PSP_CHANNEL_GREEN:
	      return PSP_CHANNELGREEN;
	  case PSP_CHANNEL_BLUE:
	      return PSP_CHANNELBLUE;
	  default:
	      PSPErrno = PSPENOPSP;
	      return PSP_CHANNELNONE;
	}
    }
    else if (chinfo.channel_type == PSP_CHANNEL_COMPOSITE)
    {
	switch (chinfo.bitmap_type)
	{
	  case PSP_DIB_TRANS_MASK:
	      return PSP_CHANNELTRANSMASK;
	  case PSP_DIB_USER_MASK:
	      return PSP_CHANNELUSERMASK;
	  case PSP_DIB_SELECTION:
	      return PSP_CHANNELSELECTION;
	  case PSP_DIB_ALPHA_MASK:
	      return PSP_CHANNELALPHA;
	  default:
	      PSPErrno = PSPENOPSP;
	      return PSP_CHANNELNONE;
	}
    }
    
    PSPErrno = PSPENOPSP;
    return PSP_CHANNELNONE;
}

int
PSPGetChannelField (PSP * psp,
		    pspf_t field,
		    void * value)
{
    pspcht_t chtype;
    
    PSPErrno = PSPENOERR;

    switch (field)
    {
      case PSPF_CHANNELTYPE:
	  if ((chtype = PSPGetChannelType (psp)) == PSP_CHANNELNONE)
	      return 0;
	  *((pspcht_t *) value) = chtype;
	  break;
      case PSPF_CHANNELNUMBER:
	  *((pspcht_t *) value) = psp->channelnum;
	  break;
      case PSPF_CHANNELBITMAPTYPE:
	  *((pspword_t *) value) =
	      psp->layer[psp->layernum]->channel[psp->channelnum]->
	      info.bitmap_type;
	  break;
      case PSPF_CHANNELCHANNELTYPE:
	  *((pspword_t *) value) =
	      psp->layer[psp->layernum]->channel[psp->channelnum]->
	      info.channel_type;
	  break;
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }
    return 1;
}

int
PSPReadScanline (PSP * psp,
		 pspbyte_t * data,
		 pspdword_t row)
{
    unsigned long col;
    struct PSPChannelInfo * channel;
    struct PSPLayerInfo * layer;
    struct PSPAlphaInfo * alpha;
    pspdword_t scanlinesize;
    pspdword_t scanlinealign;
#if 0
    unsigned char filler[4];
#endif

    PSPErrno = PSPENOERR;

    if (psp->inalpha)
    {
	if (psp->alpha == NULL)
	{
	    PSPErrno = PSPEINVAL;
	    return 0;
	}
	    /* Check if layer out of range */
	if ((psp->layernum < 0) || (psp->layernum >= psp->alphacount))
	{
	    debug (16, "Alpha channel out of valid range\n");
	    PSPErrno = PSPEINVAL;
	    return 0;
	}
	alpha = psp->alpha[psp->layernum];
    
	    /* Same for channel */
	if ((psp->channelnum < 0) || (psp->channelnum >= alpha->info.channels))
	{
	    debug (16, "Alpha subchannel out of valid range\n");
	    PSPErrno = PSPEINVAL;
	    return 0;
	}
	channel = alpha->channel[psp->channelnum];
    
	scanlinesize = alpha->scanlinebytes;
	scanlinealign = alpha->scanlinealign;
    }
    else if (psp->inselection)
    {
	    /* Check if selection channel ok */
	if ((psp->channelnum < 0)
	    || (psp->channelnum >= psp->selection->channels))
	{
	    debug (16, "Selection channel out of valid range\n");
	    PSPErrno = PSPEINVAL;
	    return 0;
	}
	channel = psp->selection->channel[psp->channelnum];
    
	scanlinesize = psp->selection->scanlinebytes;
	scanlinealign = psp->selection->scanlinealign;
    }
    else
    {
	    /* Check if layer out of range */
	if ((psp->layernum < 0) || (psp->layernum >= psp->general.layers))
	{
	    debug (16, "Layer out of valid range\n");
	    PSPErrno = PSPEINVAL;
	    return 0;
	}
	layer = psp->layer[psp->layernum];
    
	    /* Same for channel */
	if ((psp->channelnum < 0) || (psp->channelnum >= layer->info.channels))
	{
	    debug (16, "Layer channel out of valid range\n");
	    PSPErrno = PSPEINVAL;
	    return 0;
	}
	channel = layer->channel[psp->channelnum];
    
	scanlinesize = layer->scanlinebytes;
	scanlinealign = layer->scanlinealign;
    }

    if (psp->general.compression_type == PSP_COMP_NONE)
    {
	if (PSPread_int8 (psp->fp, data, scanlinesize) != scanlinesize)
	{
	    debug (16, "Error reading channel data!\n");
	    return 0;
	}
#if 0
	if (PSPread_int8 (psp->fp, filler, scanlinealign - scanlinesize) !=
	    (scanlinealign - scanlinesize))
	{
	    debug (16, "Error reading channel data!\n");
	    return 0;
	}
#endif
    }
    else if (psp->general.compression_type == PSP_COMP_RLE)
    {
	int runtype = psp->rle.runtype;
	unsigned char runcount = psp->rle.runcount;
	unsigned char lastread = psp->rle.lastread;
#if !RLESLOW
	int inchar;
#endif
	char * d;

	d = data;

	for (col = 0; col < scanlinesize; col++, runcount--)
	{
	    if (runcount <= 0)
	    {
#if RLESLOW
		if (PSPread_int8 (psp->fp, &runcount, 1) != 1)
#else
		if ((inchar = getc(psp->fp)) == EOF)
#endif
		{
		    debug (16, "Error reading channel data!\n");
		    return 0;
		}
#if !RLESLOW
		runcount = (unsigned int) inchar;
#endif
		if (runcount > 128)
		{
		    runtype = 1;
		    runcount -= 128;
#if RLESLOW
		    if (PSPread_int8 (psp->fp, &lastread, 1) != 1)
#else
		    if ((inchar = getc(psp->fp)) == EOF)
#endif
		    {
			debug (16, "Error reading channel data!\n");
			return 0;
		    }
#if !RLESLOW
		    lastread = (unsigned char) inchar;
#endif
		}
		else
		    runtype = 0;
	    }
	    if (runtype == 0)
	    {
		if (col < scanlinesize)
		{
#if RLESLOW		    
		    if (PSPread_int8 (psp->fp, d++, 1) != 1)
#else
		    if ((inchar = getc(psp->fp)) == EOF)
#endif
		    {
			debug (16, "Error reading channel data!\n");
			return 0;
		    }
#if ! RLESLOW
		    lastread = (unsigned char) inchar;
		    *d++ = (unsigned char) lastread;
#endif
		}
#if 0
		else
		{
#if RLESLOW		    
		    if (PSPread_int8 (psp->fp, &lastread, 1) != 1)
#else
		    if ((inchar = getc(psp->fp)) == EOF)
#endif
		    {
			debug (16, "Error reading channel data!\n");
			return 0;
		    }
#if !RLESLOW
		    lastread = inchar;
#endif
		}
#endif
	    }
	    else
		if (col < scanlinesize)
		    *d++ = lastread;
	}

	    /* Remember this and that until next time */
	psp->rle.runtype = runtype;
	psp->rle.runcount = runcount;
	psp->rle.lastread = lastread;
    }
    else
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    return 1;
}
