/****************************************************************
 * Paint Shop Pro File Format Library
 * Creator information handling
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: pspcreator.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: pspcreator.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#define DEBUGENV psplibdbg

#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include <psperrno.h>
int PSPErrno;

static char * PSP_CreatorAppName [4] =
{
    "Unknown application",
    "Paint Shop Pro",
    "libpsp",
    "Gimp",
};

struct PSPCreatorInfo *
PSPReadCreatorInfo (FILE * fp, DWORD firstsize, DWORD blocksize,
		    struct PSPCreatorInfo * creator)
{
    fpos_t endpos;
    fpos_t nextpos;
    fpos_t pos;
    B_ARRAY signature[4];
    WORD id;
    DWORD size;
#if DEBUG
#define MAXCREATORSTRINGSIZE 256
    B_ARRAY string[MAXCREATORSTRINGSIZE];
    DWORD dword = 0;
#if APPVERENABLED
    WORD appmjr, appmnr;
#endif /* APPVERENABLED */
    time_t psptime = 0;
#endif /* DEBUG */

    if (fgetpos (fp, &creator->filepos) != 0)
    {
	return NULL;
    }
    endpos = creator->filepos + blocksize;
    
    for (creator->count = 0; ; creator->count++)
    {
	if (fgetpos (fp, &pos) != 0)
	    return NULL;
	if (pos >= endpos)
	{
	    debug (32, "PSP Creator number of fields: %d\n", creator->count);
	    return creator;
	}
	
	if (PSPread_int8 (fp, signature, sizeof (signature))
	    != sizeof (signature)
	    || PSPread_int16 (fp, &id, 1) != 1
	    || PSPread_int32 (fp, &size, 1) != 1
	    )
	{
	    if (! feof (fp))
		return NULL;
	    debug (16, "Error reading next creator chunk (short file)!\n");
	    return NULL;
	}
	
	if (memcmp (PSP_SubBlockSig, signature, sizeof PSP_SubBlockSig) != 0)
	{
	    debug (16, "No creator data field signature found where one "
		   "should be! (%c%c%c%c)\n", signature[0],
		   signature[1], signature[2], signature[3]);
	    PSPErrno = PSPENOPSP;
	    return NULL;
	}

	if (fgetpos (fp, &nextpos) != 0)
	    return NULL;
	
	nextpos += size;
	
	debug (64, "  id: %hu, size: %lu (->%ld)\n", id, size, nextpos);

#if DEBUG
	debug (32, "PSP Creator Keyword: ");

	switch (id)
	{
	  case PSP_CRTR_FLD_TITLE:
	      debug (32, "Title\n");
	      break;
	  case PSP_CRTR_FLD_CRT_DATE:
	      debug (32, "Creation Date\n");
	      break;
	  case PSP_CRTR_FLD_MOD_DATE:
	      debug (32, "Modification Date\n");
	      break;
	  case PSP_CRTR_FLD_ARTIST:
	      debug (32, "Artist\n");
	      break;
	  case PSP_CRTR_FLD_CPYRGHT:
	      debug (32, "Copyright\n");
	      break;
	  case PSP_CRTR_FLD_DESC:
	      debug (32, "Description\n");
	      break;
	  case PSP_CRTR_FLD_APP_ID:
	      debug (32, "Application ID\n");
	      break;
	  case PSP_CRTR_FLD_APP_VER:
	      debug (32, "Application Version\n");
	      break;
#endif
	}

	debug (32, "PSP Creator Value: ");

	switch (id)
	{      case PSPF_CREATORTITLE:

	  case PSP_CRTR_FLD_TITLE:
	  case PSP_CRTR_FLD_ARTIST:
	  case PSP_CRTR_FLD_CPYRGHT:
	  case PSP_CRTR_FLD_DESC:
	      if ((dword = PSPread_int8 (fp, string,
					 MIN ((sizeof (string) - 1), size)))
		  < MIN ((sizeof (string) - 1), size))
	      {
		  debug (16, "Read error in creator value chunk "
			 "(short file?)\n");
	      }
	      string [dword] = '\0';
	      debug (32, "%s\n", string);
	      break;
	      
	  case PSP_CRTR_FLD_CRT_DATE:
	  case PSP_CRTR_FLD_MOD_DATE:
	      if (sizeof (psptime) != 4)
	      {
		  debug (32, "Data type problem, code needs to get fixed\n");
	      }
	      else
	      {
		  if (PSPread_int32 (fp, &psptime, 1) == 1)
		      debug (32, "%s", ctime (&psptime));
	      }
	      break;
	  case PSP_CRTR_FLD_APP_ID:
	      if (PSPread_int32 (fp, &dword, 1) == 1)
	      {
		  switch (dword)
		  {
		    case PSP_CREATOR_APP_UNKNOWN:
			debug (32, "Unknown application\n");
			break;
		    case PSP_CREATOR_APP_PAINT_SHOP_PRO:
			debug (32, "Paint Shop Pro\n");
			break;
		    case PSP_CREATOR_APP_3RDPTY_LIBPSP:
			debug (32, "Third party, libpsp\n");
			break;
		    case PSP_CREATOR_APP_3RDPTY_GIMP:
			debug (32, "Third party, Gimp\n");
			break;
		    default:
			debug (32, "Application not known yet. "
			       "Code needs help\n");
		  }
	      }
	      break;

/* Information from JASC Software, Inc.:
 * We are planning just to have two numbers to represent the version.
 * Thus, decode the first WORD as a number, and the second WORD as a
 * number and you will get our total version number.  I do not see an
 * exact match in the testdata, so this is disabled at the moment.
 */
	  case PSP_CRTR_FLD_APP_VER:
#if APPVERENABLED
	      if (PSPread_int32 (fp, &dword, 1) == 1)
		  debug (32, "%08x\n", dword);
#else
	      if (PSPread_int32 (fp, &dword, 1) == 1)
		  debug (32, "This is not supported at the moment!\n");
#endif /* APPVERENABLED */
	      break;
	}
	if (fsetpos (fp, &nextpos) != 0)
	{
	    debug (16, "File positioning error\n");
	    return NULL;
	}
    }

    debug (32, "PSP Creator number of fields: %d\n", creator->count);

    return creator;
}

/****************************************************************
 * GET FIELDS Interface Routine
 ****************************************************************/

/****************************************************************
 * PSPGetCreatorField
 *
 * Get information from the General Image Block
 */
static long
PSPCreatorPosField (PSP * psp,
		    WORD id)
{
    fpos_t pos;
    DWORD size;
    int i;
    WORD chunkid;

	/* go directly to the first creator block and skip the signature
	 */
    pos = psp->creator.filepos + 4;
    if (fsetpos (psp->fp, &pos) != 0)
	return -1;

    debug (128, "Wanted id: %hu\n", id);
    
    debug (64, "New file position is %ld\n", (long) pos);
    
    if (PSPread_int16 (psp->fp, &chunkid, 1) != 1
	|| PSPread_int32 (psp->fp, &size, 1) != 1)
	return -1;
    i = 0;

    while ((chunkid != id) && (i < psp->creator.count))
    {
	pos += size + 10;
	if (fsetpos (psp->fp, &pos) != 0)
	    return -1;

	debug (64, "New file position is %ld (size: %ld)\n", (long) pos,
	       (long) size);

	i++;
	if (PSPread_int16 (psp->fp, &chunkid, 1) != 1
	    || PSPread_int32 (psp->fp, &size, 1) != 1)
	    return -1;
    }
    if (chunkid == id)
	return (long) size;
    else
	return 0;
}

int
PSPGetCreatorField (PSP * psp,
		    pspf_t field,
		    void * value)
{
    long size;
    WORD id;

	     
    switch (field)
    {
      case PSPF_CREATORTITLE:
	  id = PSP_CRTR_FLD_TITLE;
	  break;
      case PSPF_CREATORCDATE:
	  id = PSP_CRTR_FLD_CRT_DATE;
	  break;
      case PSPF_CREATORARTIST:
	  id = PSP_CRTR_FLD_ARTIST;
	  break;
      case PSPF_CREATORCOPYRIGHT:
	  id = PSP_CRTR_FLD_CPYRGHT;
	  break;
      case PSPF_CREATORDESC:
	  id = PSP_CRTR_FLD_DESC;
	  break;
      case PSPF_CREATORAPPID:
	  id = PSP_CRTR_FLD_APP_ID;
	  break;
#if APPVERENABLED
      case PSPF_CREATORAPPVER:
	  id = PSP_CRTR_FLD_APP_VER;
	  break;
#endif /* APPVERENABLED */
      case PSPF_CREATORAPP:
	  id = PSP_CRTR_FLD_APP_ID;
	  break;
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }

    if ((size = PSPCreatorPosField (psp, id)) < 0)
	return 0;

    switch (field)
    {
      case PSPF_CREATORTITLE:
      case PSPF_CREATORARTIST:
      case PSPF_CREATORCOPYRIGHT:
      case PSPF_CREATORDESC:
	      /* Make it a C String with the nullbyte at the end */
	      /* This is exportable memory so use the PSP wrappers */
	  if ((*((char **) value) = PSPMalloc (size + 1)) == NULL)
	      return 0;
	  (*((char **)value))[size] = '\0';
	  if (PSPread_int8 (psp->fp, *((char **) value), size) != size)
	  {
	      PSPFree (*((char **)value));
	      return 0;
	  }
	  break;
      case PSPF_CREATORCDATE:
	  if (size != sizeof (time_t))
	      return 0;
	  if (PSPread_time_t (psp->fp, (time_t *) value, 1) != 1)
	      return 0;
	  break;
      case PSPF_CREATORAPPID:
	  if (size != 4)
	      return 0;
	  if (PSPread_int32 (psp->fp, (DWORD *) value, 1) != 1)
	      return 0;
	  break;
#if APPVERENABLED
      case PSPF_CREATORAPPVER:
	  if (size != 4)
	      return 0;
	  {
	      WORD major, minor;

	      if (PSPread_int16 (psp->fp, &minor, 1) != 1
		  || PSPread_int16 (psp->fp, &major, 1) != 1)
		  return 0;
	      ((pspword_t *) value)[0] = major;
	      ((pspword_t *) value)[1] = minor;
	  }
	  break;
#endif /* APPVERENABLED */
      case PSPF_CREATORAPP:
	  if (size != 4)
	      return 0;
	  {
	      pspword_t minor, major;
	      pspdword_t appid;
	      unsigned char tmp[128];

	      if (PSPread_int32 (psp->fp, &appid, 1) != 1)
		  return 0;
	      if ((size = PSPCreatorPosField (psp, PSP_CRTR_FLD_APP_VER)) < 0)
		  return 0;
	      if (size != 4)
		  return 0;
	      if (PSPread_int16 (psp->fp, &minor, 1) != 1
		  || PSPread_int16 (psp->fp, &major, 1) != 1)
		  return 0;

#if APPVERENABLED
	      sprintf (tmp, "%s 0x%04x%04x", PSP_CreatorAppName[appid],
		       major, minor);
#else
	      sprintf (tmp, "%s", PSP_CreatorAppName[appid]);
#endif /* APPVERENABLED */
	      if ((*((char **) value) = PSPMalloc (strlen (tmp) + 1)) == NULL)
		  return 0;

	      strcpy (*((char **)value), tmp);
	  }
	  break;
      default:
	  return 0;
    }
    return 1;
}
