/****************************************************************
 * Paint Shop Pro File Format Library
 * Routines dealing with General image information
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: pspgeneral.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: pspgeneral.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#define DEBUGENV psplibdbg
#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include <psperrno.h>
int PSPErrno;

/****************************************************************
 * ROUTINES to read the PSP structure when the file is opened
 * for reading
 ****************************************************************/
 
/* This is not 100 % compatible to the way the other block attribute
 * chunks are read.
 */

struct PSP_GeneralImageAttrInfoChunk *
PSPReadGeneralImageAttrInfo (FILE * fp,
			     struct PSP_GeneralImageAttrInfoChunk * g)
{    
    if (PSPread_int32 (fp, &g->width, 1) != 1
	|| PSPread_int32 (fp, &g->height, 1) != 1
	|| PSPread_double (fp, &g->res_value, 1) != 1
	|| PSPread_int8 (fp, &g->res_metric, 1) != 1
	|| PSPread_int16 (fp, &g->compression_type, 1) != 1
	|| PSPread_int16 (fp, &g->bits, 1) != 1
	|| PSPread_int16 (fp, &g->planes, 1) != 1
	|| PSPread_int32 (fp, &g->colors, 1) != 1
	|| PSPread_int8 (fp, &g->isgreyscale, 1) != 1
	|| PSPread_int32 (fp, &g->total_size, 1) != 1
	|| PSPread_int32 (fp, &g->active_layer, 1) != 1
	|| PSPread_int16 (fp, &g->layers, 1) != 1
	)
    {
	return NULL;
    }
    
    debug (32, "PSP General Image Attr . width: %lu\n",
	   g->width);
    debug (32, "PSP General Image Attr . height: %lu\n",
	   g->height);
#if X86_DOUBLE_IS_OK
    debug (32, "PSP General Image Attr . res_value: %f\n",
	   g->res_value);
#else /* X86_DOUBLE_IS_OK */
    debug (32, "PSP General Image Attr . res_value UNSUPPORTED -> "
	   "default: %f\n", g->res_value);
#endif /* X86_DOUBLE_IS_OK */
    debug (32, "PSP General Image Attr . res_metric: ");
    switch (g->res_metric)
    {
      case PSP_METRIC_UNDEFINED:
	  debug (32, "undefined\n");
	  break;
      case PSP_METRIC_INCH:
	  debug (32, "inch\n");
	  break;
      case PSP_METRIC_CM:
	  debug (32, "cm\n");
	  break;
      default:
	  debug (32, "unsupported value\n");
    }
    debug (32, "PSP General Image Attr . compression_type: ");
    switch (g->compression_type)
    {
      case PSP_COMP_NONE:
	  debug (32, "uncompressed\n");
	  break;
      case PSP_COMP_RLE:
	  debug (32, "RLE\n");
	  break;
      case PSP_COMP_LZ77:
	  debug (32, "LZ77\n");
	  break;
      default:
	  debug (32, "unsupported value\n");
    }
    debug (32, "PSP General Image Attr . bits: %hu\n",
	   g->bits);
    debug (32, "PSP General Image Attr . planes: %hu\n",
	   g->planes);
    debug (32, "PSP General Image Attr . colors: %lu\n",
	   g->colors);
    debug (32, "PSP General Image Attr . isgreyscale: %s\n",
	   ((g->isgreyscale == 0) ? "No" : "Yes"));
    debug (32, "PSP General Image Attr . total_size: %lu\n",
	   g->total_size);
    debug (32, "PSP General Image Attr . active_layer: %ld\n",
	   g->active_layer);
    debug (32, "PSP General Image Attr . layers: %hu\n",
	   g->layers);

    return g;
}

/****************************************************************
 * GET FIELDS Interface Routines for various PSP Blocks
 ****************************************************************/

/****************************************************************
 * PSPGetGeneralField
 *
 * Get information from the General Image Block
 */
int
PSPGetGeneralField (PSP * psp,
		    pspf_t field,
		    void * value)
{
    switch (field)
    {
      case PSPF_IMAGEWIDTH:
	  *((long *) value) = psp->general.width;
	  break;
      case PSPF_IMAGEHEIGHT:
	  *((long *) value) = psp->general.height;
	  break;
      case PSPF_IMAGERESVALUE:
	  *((double *) value) = psp->general.res_value;
#if ! X86_DOUBLE_IS_OK
	  PSPErrno = PSPENIMPL;
	  return 0;
#endif
	  break;
      case PSPF_IMAGERESUNIT:
	  *((pspenum_t *) value) = (pspenum_t) psp->general.res_metric;
	  break;
      case PSPF_IMAGECOMPRTYPE:
	  *((int *) value) = (pspenum_t) psp->general.compression_type;
	  break;
      case PSPF_IMAGEBITDEPTH:
	  *((unsigned short *) value) = psp->general.bits;
	  break;
      case PSPF_IMAGECOLORCOUNT:
	  *((unsigned long *) value) = psp->general.colors;
	  break;
      case PSPF_IMAGEISGREYSCALE:
	  *((pspbool_t *) value) = (pspbool_t) psp->general.isgreyscale;
	  break;
      case PSPF_IMAGEACTIVELAYER:
	  *((long *) value) = psp->general.active_layer;
	  break;
      case PSPF_IMAGELAYERCOUNT:
	  *((long *) value) = (long) psp->general.layers;
	  break;
      default:
	  return 0;
    }
    
    return 1;
}
