/****************************************************************
 * Paint Shop Pro File Format Library
 * Layer specific routines
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: psplayer.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: psplayer.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#define DEBUGENV psplibdbg
#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include <psperrno.h>
int PSPErrno;


struct PSPLayerInfo **
PSPReadLayerInfo (PSP * psp, DWORD firstsize, DWORD blocksize)
{
    fpos_t endpos;
    fpos_t nextpos;
    fpos_t pos;
    int i, j;
    struct PSPLayerInfo * this;
    struct PSPLayerInfo ** layer;
    FILE * fp = psp->fp;
    int pixelsperlong;		/* Number of bits per pixel in one */
    int pixelsperbyte;		/* channel */
    B_ARRAY signature[4];
    WORD id;
    DWORD size;
    DWORD subblocksize;

    if (fgetpos (fp, &endpos) != 0)
	return NULL;
    endpos += blocksize;

	/* calloc clears memory, which we use to our advantage later */
    if ((layer = PSPLIBCalloc (psp->general.layers,
			       sizeof (struct PSPLayerInfo *)))
	== NULL)
    {
	debug (8, "Unable to allocate memory for Layer Info Array\n");
	return NULL;
    }

	/* 24 bit image type has 3 channels, so this is the result:
	 * bits/pixel  bits/channel  pixels/long
	 *     1           1            32
	 *     4           4             8
	 *     8           8             4
	 *    24           8             4!!!
	 */

    pixelsperlong = 32 / MIN (psp->general.bits, 8);
    pixelsperbyte = 8 / MIN (psp->general.bits, 8);
	
    for (i = 0; i < psp->general.layers ; i++)
    {
	if (fgetpos (fp, &pos) != 0)
	{
	    debug (16, "Either some error occured in fgetpos or the "
		   "end-of-file has been reached before all data has "
		   "been read");
	    return NULL;
	}
	
	if (pos >= endpos)
	{
	    debug (16, "Something wrong here, not all layers read but"
		   "end of block already reached!\n");
	    return NULL;
	}
	
	if (PSPread_int8 (fp, signature, sizeof (signature))
	    != sizeof (signature)
	    || PSPread_int16 (fp, &id, 1) != 1
	    || PSPread_int32 (fp, &size, 1) != 1
	    || PSPread_int32 (fp, &subblocksize, 1) != 1
	    )
	{
	    if (! feof (fp))
		return NULL;
	    debug (16, "Error reading next layer sub-block (short file)!\n");
	    return NULL;
	}
	
	if (memcmp (PSP_BlockSig, signature, sizeof (PSP_BlockSig)) != 0)
	{
	    debug (16, "No layer sub-block signature found where expected! "
		   "(%c%c%c%c)\n", signature[0],
		   signature[1], signature[2], signature[3]);
	    PSPErrno = PSPENOPSP;
	    return NULL;
	}

	if (id != PSP_LAYER_BLOCK)
	{
	    debug (16, "Wrong block identifier %hu where PSP_LAYER_BLOCK "
		   "(%hu) expected!\n", id, PSP_LAYER_BLOCK);
	    return NULL;
	}

	if (fgetpos (fp, &nextpos) != 0)
	    return NULL;
	    
	nextpos += subblocksize;

	debug (64, "  layer: %d, size: %lu (->%ld)\n", i, size, nextpos);

	if ((layer[i] = PSPLIBMalloc (sizeof (struct PSPLayerInfo))) == NULL)
	{
	    debug (8, "Unable to allocate memory for Layer %d Info \n", i);
	    return NULL;
	}

	this = layer[i];
	
	if (PSPread_int8 (fp, this->info.name, PSP_LAYERINFONAMELEN) !=
	    PSP_LAYERINFONAMELEN
	    || PSPread_int8 (fp, &this->info.type, 1) != 1
	    || PSPread_int32 (fp, &this->info.rect.xmin, 1) != 1
	    || PSPread_int32 (fp, &this->info.rect.ymin, 1) != 1
	    || PSPread_int32 (fp, &this->info.rect.xmax, 1) != 1
	    || PSPread_int32 (fp, &this->info.rect.ymax, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_rect.xmin, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_rect.ymin, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_rect.xmax, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_rect.ymax, 1) != 1
	    || PSPread_int8 (fp, &this->info.opacity, 1) != 1
	    || PSPread_int8 (fp, &this->info.blending_mode, 1) != 1
	    || PSPread_int8 (fp, &this->info.visible, 1) != 1
	    || PSPread_int8 (fp, &this->info.protect_trans, 1) != 1
	    || PSPread_int8 (fp, &this->info.link_group, 1) != 1
	    || PSPread_int32 (fp, &this->info.mask_rect.xmin, 1) != 1
	    || PSPread_int32 (fp, &this->info.mask_rect.ymin, 1) != 1
	    || PSPread_int32 (fp, &this->info.mask_rect.xmax, 1) != 1
	    || PSPread_int32 (fp, &this->info.mask_rect.ymax, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_mask_rect.xmin, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_mask_rect.ymin, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_mask_rect.xmax, 1) != 1
	    || PSPread_int32 (fp, &this->info.saved_mask_rect.ymax, 1) != 1
	    || PSPread_int8 (fp, &this->info.mask_linked, 1) != 1
	    || PSPread_int8 (fp, &this->info.mask_disabled, 1) != 1
	    || PSPread_int8 (fp, &this->info.inv_mask_blend, 1) != 1
	    || PSPread_int16 (fp, &this->info.blend_ranges, 1) != 1)
	{
	    debug (16, "Error reading layer information\n");
	    return NULL;
	}

	debug (32, "Layer %d -> name: %.256s\n", i, this->info.name);
#if DEBUG
	debug (32, "Layer %d -> type: ", i);
	switch (this->info.type)
	{
	  case PSP_LAYER_NORMAL:
	      debug (32, "normal\n");
	      break;
	  case PSP_LAYER_FLOATING_SELECTION:
	      debug (32, "floating selection\n");
	      break;
	  default:
	      debug (32, "unknown, needs to get fixed!\n");
	}
#endif
	debug (32, "Layer %d -> rect: (%ld, %ld), (%ld, %ld)\n", i,
	       this->info.rect.xmin, this->info.rect.ymin,
	       this->info.rect.xmax, this->info.rect.ymax);
	debug (32, "Layer %d -> saved rect: (%ld, %ld), (%ld, %ld)\n", i,
	       this->info.saved_rect.xmin, this->info.saved_rect.ymin,
	       this->info.saved_rect.xmax, this->info.saved_rect.ymax);
	debug (32, "Layer %d -> opacity: %d\n", i, (int)
	       this->info.opacity);
#if DEBUG
	{
	    char * blending_modes[] =
	    {
		"Normal",
		"Darken",
		"Lighten",
		"Hue",
		"Saturation",
		"Color",
		"Luminosity",
		"Multiply",
		"Screen",
		"Dissolve",
		"Overlay",
		"Hard Light",
		"Soft Light",
		"Difference",
		"Dodge",
		"Burn",
		"Exclusion",
	    };

	    if (this->info.blending_mode > 16) {
		debug (32, "Layer %d -> blending mode: %d "
		       "UNKNOWN to program!\n",
		       i, (int) this->info.blending_mode);
	    }
	    else
	    {
		debug (32, "Layer %d -> blending mode: %s\n", i,
		       blending_modes[this->info.blending_mode]);
	    }
	}
#endif
	
	debug (32, "Layer %d -> visible? %s\n", i,
	       (this->info.visible ? "yes" : "no"));
	debug (32, "Layer %d -> protect transparency? %s\n", i,
	       (this->info.protect_trans ? "yes" : "no"));
	debug (32, "Layer %d -> link group: %d\n", i,
	       (int) this->info.link_group);
	debug (32, "Layer %d -> mask rect: (%ld, %ld), (%ld, %ld)\n", i,
	       this->info.mask_rect.xmin, this->info.mask_rect.ymin,
	       this->info.mask_rect.xmax, this->info.mask_rect.ymax);
	debug (32, "Layer %d -> saved mask rect: (%ld, %ld), (%ld, %ld)\n", i,
	       this->info.saved_mask_rect.xmin,
	       this->info.saved_mask_rect.ymin,
	       this->info.saved_mask_rect.xmax,
	       this->info.saved_mask_rect.ymax);
	debug (32, "Layer %d -> mask linked? %s\n", i,
	       (this->info.mask_linked ? "yes" : "no"));
	debug (32, "Layer %d -> mask disabled? %s\n", i,
	       (this->info.mask_disabled ? "yes" : "no"));
	debug (32, "Layer %d -> invert mask on blend? %s\n", i,
	       (this->info.inv_mask_blend ? "yes" : "no"));
	debug (32, "Layer %d -> number of blend ranges: %hu\n", i,
	       this->info.blend_ranges);

	if (this->info.blend_ranges > 5)
	{
	    debug (16, "Error: more than 5 blend ranges\n");
	    return (NULL);
	}
	
	for (j = 0; j < 5; j++)
	{
	    if (PSPread_int8 (fp, &this->info.blendrange[j].src.in.low, 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].src.out.low,
				 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].src.in.high,
				 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].src.out.high,
				 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].dst.in.low,
				 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].dst.out.low,
				 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].dst.in.high,
				 1) != 1
		|| PSPread_int8 (fp, &this->info.blendrange[j].dst.out.high,
				 1) != 1)
	    {
		debug (16, "Error reading blending ranges\n");
		return NULL;
	    }

	    debug (32, "Layer %d -> blendrange %d src  in  low: %d\n", i, j,
		   (int) this->info.blendrange[j].src.in.low);
	    debug (32, "Layer %d -> blendrange %d src  in high: %d\n", i, j,
		   (int) this->info.blendrange[j].src.in.high);
	    debug (32, "Layer %d -> blendrange %d src out  low: %d\n", i, j,
		   (int) this->info.blendrange[j].src.out.low);
	    debug (32, "Layer %d -> blendrange %d src out high: %d\n", i, j,
		   (int) this->info.blendrange[j].src.out.high);
	    debug (32, "Layer %d -> blendrange %d dst  in  low: %d\n", i, j,
		   (int) this->info.blendrange[j].dst.in.low);
	    debug (32, "Layer %d -> blendrange %d dst  in high: %d\n", i, j,
		   (int) this->info.blendrange[j].dst.in.high);
	    debug (32, "Layer %d -> blendrange %d dst out  low: %d\n", i, j,
		   (int) this->info.blendrange[j].dst.out.low);
	    debug (32, "Layer %d -> blendrange %d dst out high: %d\n", i, j,
		   (int) this->info.blendrange[j].dst.out.high);
	}
	
	if (PSPread_int16 (fp, &this->info.bitmaps, 1) != 1
	    || PSPread_int16 (fp, &this->info.channels, 1) != 1)
	{
	    debug (16, "Error reading layer information\n");
	    return NULL;
	}

	debug (32, "Layer %d -> number of bitmaps: %hu\n", i,
	       this->info.bitmaps);
	debug (32, "Layer %d -> number of channels: %hu\n", i,
	       this->info.channels);

	    /*
	     * Calibrate the scanline size
	     */
	    /* Scanlines always have to start on a 4 byte boundary */
	this->scanlinebytes = this->info.saved_rect.xmax
	    - this->info.saved_rect.xmin;
	this->scanlinealign = this->scanlinebytes;
	    /* The first pixel in a 4 byte piece already eats a full 4
	     * byte piece of memory
	     */
	this->scanlinealign += (pixelsperlong - 1);
	this->scanlinealign /= pixelsperlong;
	this->scanlinealign *= 4;
	this->scanlinebytes += (pixelsperbyte - 1);
	this->scanlinebytes /= pixelsperbyte;
	    /* Mask is always 8 bit "greyscale" */
	this->mask_scanlinebytes = this->info.saved_mask_rect.xmax
	    - this->info.saved_mask_rect.xmin;
	this->mask_scanlinealign = this->mask_scanlinebytes;
	this->mask_scanlinealign += 3;
	this->mask_scanlinealign &= 0xfffffffc;
	    /*
	this->mask_scanlinebytes += (pixelsperbyte - 1);
	this->mask_scanlinebytes /= pixelsperbyte;
	    */
	
	debug (32, "Layer %d -> scanline bytecount: %lu\n", i,
	       this->scanlinebytes);
	debug (32, "Layer %d -> scanline aligned bytecount: %lu\n", i,
	       this->scanlinealign);
	debug (32, "Layer %d -> mask scanline bytecount: %lu\n", i,
	       this->mask_scanlinebytes);
	debug (32, "Layer %d -> mask scanline aligned bytecount: %lu\n", i,
	       this->mask_scanlinealign);

	if (PSPReadChannelInfo (psp->fp, size, subblocksize,
				&this->channel, this->info.channels) == NULL)
	{
	    debug (16, "Problem reading Channel Block!\n");
	    return NULL;
	}
	
	if (fsetpos (fp, &nextpos) != 0)
	{
	    debug (16, "File positioning error\n");
	    return NULL;
	}
	
    }

    psp->layer = layer;
    
    return layer;
}

static int
PSPSetAlphaLayer (PSP * psp,
		  psplay_t layertype)
{
    if (psp->alpha == NULL)
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    switch (layertype)
    {
      case PSP_LAYERFIRSTALPHA:
	  psp->layernum = 0;
	  break;
      case PSP_LAYERLASTALPHA:
	  psp->layernum = psp->alphacount - 1;
	  break;
      case PSP_LAYERNEXTALPHA:
	  if (psp->layernum + 1 >= psp->alphacount)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum++;
	  break;
      case PSP_LAYERPREVALPHA:
	  if (psp->layernum <= 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum--;
	  break;	  
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }

    psp->inalpha = TRUE;
    
    return 1;
}
		    
int
PSPSetLayer (PSP * psp,
	     psplay_t layertype)
{
    int layernum;

    PSPErrno = PSPENOERR;

    if (layertype <= PSP_LAYERALPHAOFFSET)
    {
	return PSPSetAlphaLayer (psp, layertype);
    }
    
    switch (layertype)
    {
      case PSP_LAYERFIRST:
	  psp->layernum = 0;
	  break;
      case PSP_LAYERLAST:
	  psp->layernum = psp->general.layers - 1;
	  break;
      case PSP_LAYERNEXT:
	  if (psp->layernum + 1 >= psp->general.layers)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum++;
	  break;
      case PSP_LAYERPREV:
	  if (psp->layernum <= 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum--;
	  break;
      case PSP_LAYERFIRSTNORMAL:
	  psp->layernum = 0;
	  if ((psp->layernum < psp->general.layers)
	      && (psp->layer[psp->layernum]->info.type
	      != PSP_LAYER_NORMAL))
	  {
	      psp->inalpha = FALSE;
    
	      return PSPSetLayer (psp, PSP_LAYERNEXTNORMAL);
	  }
	  break;
      case PSP_LAYERLASTNORMAL:
	  psp->layernum = psp->general.layers - 1;
	  if ((psp->layernum >= 0)
	      && (psp->layer[psp->layernum]->info.type
	      != PSP_LAYER_NORMAL))
	  {
	      psp->inalpha = FALSE;
    
	      return PSPSetLayer (psp, PSP_LAYERPREVNORMAL);
	  }
	  break;
      case PSP_LAYERNEXTNORMAL:
	  layernum = psp->layernum;
	  while ((++layernum < psp->general.layers)
		 && (psp->layer[layernum]->info.type !=
		     PSP_LAYER_NORMAL))
	      ;
	  if (layernum >= psp->general.layers - 1)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum = layernum;
	  break;
      case PSP_LAYERPREVNORMAL:
	  layernum = psp->layernum;
	  while ((--layernum >= 0)
		 && (psp->layer[layernum]->info.type !=
		     PSP_LAYER_NORMAL))
	      ;
	  if (layernum < 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  
	  psp->layernum = layernum;
	  break;
      case PSP_LAYERFIRSTFLOAT:
	  psp->layernum = 0;
	  if ((psp->layernum < psp->general.layers)
	      && (psp->layer[psp->layernum]->info.type
	      != PSP_LAYER_FLOATING_SELECTION))
	  {
	      psp->inalpha = FALSE;
    
	      return PSPSetLayer (psp, PSP_LAYERNEXTFLOAT);
	  }
	  break;
      case PSP_LAYERLASTFLOAT:
	  psp->layernum = psp->general.layers - 1;
	  if ((psp->layernum >= 0)
	      && (psp->layer[psp->layernum]->info.type
	      != PSP_LAYER_FLOATING_SELECTION))
	  {
	      psp->inalpha = FALSE;
    
	      return PSPSetLayer (psp, PSP_LAYERPREVFLOAT);
	  }
	  
	  break;
      case PSP_LAYERNEXTFLOAT:
	  layernum = psp->layernum;
	  while ((++layernum < psp->general.layers)
		 && (psp->layer[layernum]->info.type !=
		     PSP_LAYER_FLOATING_SELECTION))
	      ;
	  if (layernum >= psp->general.layers - 1)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum = layernum;
	  break;
      case PSP_LAYERPREVFLOAT:
	  layernum = psp->layernum;
	  while ((--layernum >= 0)
		 && (psp->layer[layernum]->info.type !=
		     PSP_LAYER_NORMAL))
	      ;
	  if (layernum < 0)
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum = layernum;
	  break;
      default:
	  if ((layertype < 0) || (layertype >= psp->general.layers))
	  {
	      PSPErrno = PSPEINVAL;
	      return 0;
	  }
	  psp->layernum = layertype;
    }

    psp->inalpha = FALSE;
    
    return 1;
}

int
PSPNextLayer (PSP * psp)
{
    if (psp->inalpha)
	return (PSPSetAlphaLayer (psp, PSP_LAYERNEXTALPHA));
    else
	return (PSPSetLayer (psp, PSP_LAYERNEXT));
}

int
PSPGetLayerField (PSP * psp,
		  pspf_t field,
		  void * value)
{
    struct PSPLayerInfo * layer;

    if ((psp->inalpha) && (psp->alpha != NULL))
    {
	return PSPGetAlphaField (psp, field, value);
    }

    if ((psp->layernum < 0) || (psp->layernum >= psp->general.layers))
    {
	PSPErrno = PSPEINVAL;
	return 0;
    }

    layer = psp->layer[psp->layernum];
    
    switch (field)
    {
      case PSPF_LAYERNAME:
	  if ((*((char **) value) =
	       PSPMalloc (strlen (layer->info.name) + 1)) == NULL)
	      return 0;
	  strcpy (*((char **) value), layer->info.name);
	  break;
      case PSPF_LAYERTYPE:
	  *((psplay_t *) value) = (psplay_t) layer->info.type;
	  break;
      case PSPF_LAYEREXTENT:
	  *((psprect_t *) value) = layer->info.rect;
	  break;
      case PSPF_LAYERDATAEXTENT:
	  *((psprect_t *) value) = layer->info.saved_rect;
	  break;
      case PSPF_LAYEROPACITY:
	  *((pspbyte_t *) value) = layer->info.opacity;
	  break;
      case PSPF_LAYERMODE:
	  *((pspenum_t *) value) = (pspenum_t) layer->info.blending_mode;
	  break;
      case PSPF_LAYERVISIBLE:
	  *((pspbool_t *) value) = (pspbool_t) layer->info.visible;
	  break;
      case PSPF_LAYERTRANSPROT:
	  *((pspbool_t *) value) = (pspbool_t) layer->info.protect_trans;
	  break;
      case PSPF_LAYERLINKGROUP:
	  *((pspword_t *) value) = (pspword_t) layer->info.link_group;
	  break;
      case PSPF_LAYERMASKEXTENT:
	  *((psprect_t *) value) = layer->info.mask_rect;
	  break;
      case PSPF_LAYERMASKDATAEXTENT:
	  *((psprect_t *) value) = layer->info.saved_mask_rect;
	  break;
      case PSPF_LAYERMASKLINKED:
	  *((pspbool_t *) value) = (pspbool_t) layer->info.mask_linked;
	  break;
      case PSPF_LAYERMASKDISABLED:
	  *((pspbool_t *) value) = (pspbool_t) layer->info.mask_disabled;
	  break;
      case PSPF_LAYERMASKINVERT:
	  *((pspbool_t *) value) = (pspbool_t) layer->info.inv_mask_blend;
	  break;
      case PSPF_LAYERBITMAPCOUNT:
	  *((pspword_t *) value) = layer->info.bitmaps;
	  break;
      case PSPF_LAYERCHANNELCOUNT:
	  *((pspword_t *) value) = layer->info.channels;
	  break;
      case PSPF_LAYERNUMBER:
	  *((psplay_t *) value) = psp->layernum;
	  break;
      case PSPF_LAYERSCANLINEBYTES:
	  *((pspdword_t *) value) = layer->scanlinebytes;
	  break;
      case PSPF_LAYERSCANLINEALIGN:
	  *((pspdword_t *) value) = layer->scanlinealign;
	  break;
      case PSPF_LAYERMASKSCANLINEBYTES:
	  *((pspdword_t *) value) = layer->mask_scanlinebytes;
	  break;
      case PSPF_LAYERMASKSCANLINEALIGN:
	  *((pspdword_t *) value) = layer->mask_scanlinealign;
	  break;
      default:
	  PSPErrno = PSPEINVAL;
	  return 0;
    }
    return 1;
}
