/****************************************************************
 * Paint Shop Pro File Format Library
 * Misc. functions
 *
 * Paint Shop Pro by Jasc Software, Inc. (http://www.jasc.com/)
 *
 * $Id: pspmisc.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $
 *
 * Copyright (c) 1999
 *               Thomas Linder
 *
 * Based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This is part of libpsp and is free software; you can redistribute
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid [] = "$Id: pspmisc.c,v 1.1.1.1 1999/08/15 19:54:42 tl Exp $";

#include "psp.h"
#include <pspproto.h>

#define DEBUGENV psplibdbg
#include <debug.h>

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>

#include <psperrno.h>
int PSPErrno;


static PSP *
PSPReadHeaders (PSP * psp)
{
    FILE * fp = psp->fp;
    struct PSP_BlockHeader bh;
    fpos_t offset;

    while (!feof (fp))
    {
	if (PSPread_int8 (fp, bh.signature, sizeof (bh.signature))
	    != sizeof (bh.signature)
	    || PSPread_int16 (fp, &bh.id, 1) != 1
	    || PSPread_int32 (fp, &bh.first_size, 1) != 1
	    || PSPread_int32 (fp, &bh.size, 1) != 1)
	{
	    if (! feof (fp))
		return NULL;
	    break;
	}

	if (memcmp (PSP_BlockSig, bh.signature, sizeof PSP_BlockSig) != 0)
	{
	    PSPErrno = PSPENOPSP;
	    debug (16, "No PSP Block signature!\n");
	    return NULL;
	}

	if (fgetpos (fp, &offset) != 0)
	    return NULL;
	
	offset += bh.size;
	debug (64, "id: %hu, size: %lu (->%ld)\n", bh.id, bh.size, offset);
	
	    /* Depending on the Block Header ID call the right function
	     * to read the attribute data
	     */
	switch (bh.id)
	{
	  case PSP_CREATOR_BLOCK:
	      debug (32, "Creator Block found!\n");
	      if (PSPReadCreatorInfo (psp->fp, bh.first_size,
				      bh.size, &psp->creator)
		  == NULL)
	      {
		  return NULL;
	      }
	      break;
	  case PSP_COLOR_BLOCK:
	      if (PSPReadCMapInfo (psp->fp, &psp->palette, bh.size)
		  == NULL)
	      {
		  debug (16, "Problem reading Color Block\n");
		  return NULL;
	      }
	      break;
	  case PSP_LAYER_START_BLOCK:
	      debug (32, "Layer Start Block found!\n");
	      if (PSPReadLayerInfo (psp, bh.first_size, bh.size)
		  == NULL)
	      {
		  debug (16, "Problem reading Layer Block\n");
		  return NULL;
	      }	      
	      break;
	  case PSP_SELECTION_BLOCK:
	      debug (32, "Selection Block found!\n");
	      if (PSPReadSelectionInfo (psp->fp, &psp->selection, bh.size)
		  == NULL)
	      {
		  debug (16, "Problem reading Selection Block\n");
		  return NULL;
	      }
	      break;
	  case PSP_ALPHA_BANK_BLOCK:
	      debug (32, "Alpha Channel Block found!\n");
	      if (PSPReadAlphaInfo (psp, bh.first_size, bh.size) == NULL)
	      {
		  debug (16, "Problem reading Alpha Channel Block\n");
		  return NULL;
	      }
	      break;
	  case PSP_THUMBNAIL_BLOCK:
	      debug (16, "Thumbnail Block not implemented yet\n");
	      break;
	  case PSP_EXTENDED_DATA_BLOCK:
	      debug (16, "Extended Data Block not implemented yet\n");
	      break;
	  case PSP_TUBE_BLOCK:
	      debug (16, "Tube Block not implemented\n");
	      break;
	  default:
	      debug (16, "Unknown Block (id = %hu)\n", bh.id);
	}
	if (fsetpos (fp, &offset) != 0)
	{
	    debug (4, "Unable to seek within file\n");
	    return NULL;
	}
	debug (64, "At next header position\n");
    }

    debug (64, "No more headers, reached eof\n");
    
    return psp;
}

PSP *
PSPreadinfo (PSP * psp)
{
    struct PSP_FileHeader fh;
    struct PSP_BlockHeader bh;

    if (PSPread_int8 (psp->fp, fh.signature, sizeof (fh.signature))
	!= sizeof (fh.signature)
	|| PSPread_int16 (psp->fp, &fh.major, 1) != 1
	|| PSPread_int16 (psp->fp, &fh.minor, 1) != 1)
    {
	return NULL;
    }
    
	/* Refuse to open file if it has no valid signature */
    if (memcmp (PSP_Signature, fh.signature, sizeof PSP_Signature) != 0)
    {
	debug (16, "Oh no, %s has size %d and is not equal to %s\n",
	       PSP_Signature, sizeof PSP_Signature, fh.signature);
	debug (16, "PSP signature not found!\n");
	PSPErrno = PSPENOPSP;
	return NULL;
    }
    
    debug (32, "PSP version %hu.%hu\n", fh.major, fh.minor);
    
	/* Is this version supported? */
    if ((fh.major != 3) && (fh.minor != 0))
    {
	PSPErrno = PSPEUNSUP;
	return NULL;
    }

	/* Now try to read more image info */
    if (PSPread_int8 (psp->fp, bh.signature, sizeof (bh.signature))
	!= sizeof (bh.signature)
	|| PSPread_int16 (psp->fp, &bh.id, 1) != 1
	|| PSPread_int32 (psp->fp, &bh.first_size, 1) != 1
	|| PSPread_int32 (psp->fp, &bh.size, 1) != 1)
    {
	return NULL;
    }
    
    if (memcmp (PSP_BlockSig, bh.signature, sizeof PSP_BlockSig) != 0)
    {
	debug (16, "PSP Block signature not found!\n");
	PSPErrno = PSPENOPSP;
	return NULL;
    }
	
    if (bh.id != PSP_IMAGE_BLOCK)
    {
	debug (16, "PSP General Image Attribute Block missing!?\n");
	return NULL;
    }
    
    debug (64, "PSP General Image Attr Header -> first length: %lu\n",
	   bh.first_size);
    debug (64, "PSP General Image Attr Header -> length: %lu\n",
	   bh.size);
    debug (64, "PSP General Image Attr Chunk struct size: %lu\n",
	   sizeof (struct PSP_GeneralImageAttrInfoChunk));
    
	/* Now read the general image info */
	/* This is mandatory information so fail if there is a problem */
    if (PSPReadGeneralImageAttrInfo (psp->fp, &psp->general) == NULL)
    {
	return NULL;
    }

    if (PSPReadHeaders (psp) == NULL)
    {
	return NULL;
    }

    return psp;
}
